/*
 * nlwifi.c - implements for cfg80211 wifi drivers
 *
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include <errno.h>
#include <sys/types.h>
#include <byteswap.h>
#include <fcntl.h>
#include <net/if.h>
#include <netlink/genl/genl.h>
#include <netlink/genl/ctrl.h>
#include <netpacket/packet.h>
#include <linux/nl80211.h>

#include "debug.h"
#include "util.h"
#include "wifi.h"
#include "wext.h"
#include "nlwifi.h"

static int ieee80211_frequency_to_channel(int freq)
{
	/* see 802.11-2007 17.3.8.3.2 and Annex J */
	if (freq == 2484)
		return 14;
	else if (freq < 2484)
		return (freq - 2407) / 5;
	else if (freq >= 4910 && freq <= 4980)
		return (freq - 4000) / 5;
	else if (freq <= 45000) /* DMG band lower limit */
		return (freq - 5000) / 5;
	else if (freq >= 58320 && freq <= 64800)
		return (freq - 56160) / 2160;
	else
		return 0;
}

static enum wifi_mode ieee80211_type_to_wifi_mode(enum nl80211_iftype type)
{
	if (type == NL80211_IFTYPE_STATION)
		return WIFI_MODE_STA;
	else if (type == NL80211_IFTYPE_AP)
		return WIFI_MODE_AP;
	else if (type == NL80211_IFTYPE_AP_VLAN)
		return WIFI_MODE_AP_VLAN;
	else if (type == NL80211_IFTYPE_MONITOR)
		return WIFI_MODE_MONITOR;
	else
		return WIFI_MODE_UNKNOWN;
}

static struct nl_sock *nlwifi_socket()
{
	struct nl_sock *nl;

	nl = nl_socket_alloc();
	if (nl == NULL)
		return NULL;

	genl_connect(nl);

	return nl;
}

void *nlwifi_alloc_msg(struct nl_sock *sk, int cmd, int flags, size_t priv)
{
	struct nl_msg *msg;
	int id;

	msg = nlmsg_alloc();
	if (!msg)
		return NULL;

	id = genl_ctrl_resolve(sk, "nl80211");
	if (id < 0) {
		nlmsg_free(msg);
		return NULL;
	}

	genlmsg_put(msg, NL_AUTO_PORT, NL_AUTO_SEQ, id, priv, flags, cmd, 0);

	return msg;
}

static int nlwifi_ack_handler(struct nl_msg *msg, void *arg)
{
	int *ret = arg;

	*ret = 0;
	return NL_STOP;
}

static int nlwifi_finish_handler(struct nl_msg *msg, void *arg)
{
	int *ret = arg;

	*ret = 0;
	return NL_SKIP;
}

static int nlwifi_err_handler(struct sockaddr_nl *nla, struct nlmsgerr *err,
								void *arg)
{
	int *ret = arg;

	*ret = err->error;
	libwifi_err("Error: %s\n", strerror(-err->error));
	return NL_STOP;
}

static int nlwifi_no_seq_check(struct nl_msg *msg, void *arg)
{
	return NL_OK;
}

static int nlwifi_send_msg(struct nl_sock *nl, struct nl_msg *msg,
			int (*cmd_resp_handler)(struct nl_msg *, void *),
			void *cmd_resp_data)
{
	struct nl_cb *cb;
	int err = 1;

	if (!msg)
		return -1;

	cb = nl_cb_alloc(NL_CB_DEFAULT);
	if (!cb)
		return -1;

	nl_socket_set_cb(nl, cb);

	nl_cb_err(cb, NL_CB_CUSTOM, nlwifi_err_handler, &err);
	nl_cb_set(cb, NL_CB_FINISH, NL_CB_CUSTOM, nlwifi_finish_handler, &err);
	nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, nlwifi_ack_handler, &err);
	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, nlwifi_no_seq_check, NULL);

	if (cmd_resp_handler)
		nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, cmd_resp_handler, cmd_resp_data);


	err = nl_send_auto(nl, msg);
	if (err < 0)
		goto out;

	while(err > 0) {
		if (nl_recvmsgs(nl, cb) < 0) {
			libwifi_dbg("nl_recvmsgs failed\n");
			break;
		}
	}
out:
	nl_cb_put(cb);
	return err;
}

static int _nlwifi_get_phyname(struct nl_msg *msg, void *resp)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	//uint32_t phyid = -1;

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	/* if (tb[NL80211_ATTR_WIPHY])
		phyid = nla_get_u32(tb[NL80211_ATTR_WIPHY]); */

	if (tb[NL80211_ATTR_WIPHY_NAME]) {
		strncpy(resp, nla_get_string(tb[NL80211_ATTR_WIPHY_NAME]), 16);
	}

	/* libwifi_dbg("Phy name: %s   id = %d\n", resp, phyid); */
	return NL_SKIP;
}

int nlwifi_get_phyname(const char *ifname, char *phyname)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_WIPHY, 0, 0);
	if (!msg)
		goto out_msg_failure;

	/* prepare cmd */
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_get_phyname, phyname);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_scan(const char *ifname, struct scan_param *p)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_TRIGGER_SCAN, 0, 0);
	if (!msg)
		goto out_msg_failure;

	/* prepare cmd */
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	if (p && p->ssid[0]) {
		struct nlattr *nest;

		nest = nla_nest_start(msg, NL80211_ATTR_SCAN_SSIDS);
		if (!nest)
			goto nla_put_failure;

		libwifi_dbg("SSID to scan: '%s'  (len = %u)\n",
					p->ssid, strlen(p->ssid));
		NLA_PUT(msg, 1, strlen(p->ssid), p->ssid);
		nla_nest_end(msg, nest);
	}

	/* send cmd */
	ret = nlwifi_send_msg(nl, msg, NULL, NULL);
	nlmsg_free(msg);
	nl_socket_free(nl);
	return ret;

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

UNUSED_FUNC static void nlwifi_parse_wpa_ie(uint8_t *iebuf, size_t len, int *nxt,
			uint32_t *auth, uint32_t *p_cipher, uint32_t *g_cipher)
{
	int ielen = iebuf[1] + 2;
	int offset = 2;
	unsigned char wpa1_oui[3] = {0x00, 0x50, 0xf2};
	unsigned char wpa2_oui[3] = {0x00, 0x0f, 0xac};
	unsigned char *wpa_oui;
	uint16_t cnt = 0;
	int i;

#ifndef bit
#define bit(_x)		(1 << (_x))
#endif

	if (ielen > len)
		ielen = len;

	switch (iebuf[0]) {
	case 0x30:
		if (ielen < 4)
			return;
		wpa_oui = wpa2_oui;
		*auth |= WPA_VERSION2;
		break;
	case 0xdd:
		wpa_oui = wpa1_oui;
		*auth |= WPA_VERSION1;
		if ((ielen < 8)
			|| (memcmp(&iebuf[offset], wpa_oui, 3) != 0)
			|| (iebuf[offset + 3] != 0x01)) {
			return;
		}
		offset += 4;
		break;
	default:
		return;
	}

	/* increment for next suites */
	(*nxt)++;
	offset += 2;

	if (ielen < (offset + 4)) {
		*auth = bit(WIFI_AKM_PSK);

		/* Short ie - assume tkip/tkip */
		*p_cipher |= bit(WIFI_CIPHER_TKIP);
		*g_cipher |= bit(WIFI_CIPHER_TKIP);
		return;
	}

	/* group cipher */
	if (!memcmp(&iebuf[offset], wpa_oui, 3)) {
		switch (iebuf[offset + 3]) {
			case WIFI_CIPHER_WEP40:
			case WIFI_CIPHER_WEP104:
			case WIFI_CIPHER_TKIP:
			case WIFI_CIPHER_CCMP:
				*g_cipher = bit(iebuf[offset + 3]);
				break;
			default:
				break;
		}
	}
	offset += 4;
	if (ielen < (offset + 2)) {
		/* no pairwise cipher or auth method.. assume tkip */
		*p_cipher = bit(WIFI_CIPHER_TKIP);
		*auth = bit(WIFI_AKM_PSK);
		return;
	}

	/* pairwise ciphers */
	cnt = iebuf[offset] | (iebuf[offset + 1] << 8);
	offset += 2;

	if (ielen < (offset + 4 * cnt))
		return;

	for (i = 0; i < cnt; i++) {
		if(!memcmp(&iebuf[offset], wpa_oui, 3)) {
			if (iebuf[offset+3] == WIFI_CIPHER_WEP40)
				*p_cipher = WIFI_CIPHER_WEP40;
			else if(iebuf[offset+3] == WIFI_CIPHER_WEP104)
				*p_cipher = WIFI_CIPHER_WEP104;
			else if(iebuf[offset+3] == WIFI_CIPHER_TKIP)
				*p_cipher = WIFI_CIPHER_TKIP;
			else if(iebuf[offset+3] == WIFI_CIPHER_CCMP)
				*p_cipher = WIFI_CIPHER_CCMP;
		}
		offset += 4;
	}

	if (ielen < (offset + 2))
		return;

	/* authentication suites */
	cnt = iebuf[offset] | (iebuf[offset + 1] << 8);
	offset += 2;

	if (ielen < (offset + 4 * cnt))
		return;

	for (i = 0; i < cnt; i++) {
		if (!memcmp(&iebuf[offset], wpa_oui, 3)) {
			switch (iebuf[offset + 3]) {
			case WIFI_AKM_DOT1X:
			case WIFI_AKM_PSK:
			case WIFI_AKM_FT_DOT1X:
			case WIFI_AKM_FT_PSK:
				*auth |= bit(iebuf[offset + 3]);
				break;
			default:
				break;
			}
		}
		offset += 4;
	}

	if (ielen < (offset + 1))
		return;
}

static int nlwifi_get_ie(uint8_t *ies, size_t len, uint8_t e, void *data)
{
	int offset = 0;
	int ielen = 0;
	uint8_t eid;

	while (offset <= (len - 2)) {
		eid = ies[offset];
		ielen = ies[offset + 1];
		if (eid != e) {
			offset += ielen + 2;
			continue;
		}

		switch (eid) {
		case 11:
			if (ielen > sizeof(struct wifi_ap_load))
				return -1;
			break;
		case 127:
			if (ielen > sizeof(struct wifi_caps_ext))
				return -1;
			break;
		case 45:
			if (ielen > sizeof(struct wifi_caps_ht))
				return -1;
			break;
		case 191:
			if (ielen > sizeof(struct wifi_caps_vht))
				return -1;
			break;
		case 70:
			if (ielen > sizeof(struct wifi_caps_rm))
				return -1;
			break;
		default:
			break;
		}

		memcpy((uint8_t *)data, &ies[offset + 2], ielen);
		return 0;
	}

	return -ENOENT;
}

static int nlwifi_get_bandwidth_from_ie(uint8_t *ies, size_t len,
						enum wifi_bw *bw)
{
	int offset = 0;
	int ielen = 0;
	uint8_t *p = NULL;
	int cntr = 0;

#ifndef bit
#define bit(n)	(1 << (n))
#endif

	while (offset <= (len - 2)) {
		ielen = ies[offset + 1];
		p = &ies[offset + 2];

		switch (ies[offset]) {
		case 61:   /* HT oper */
			{
				uint8_t *ht_op;
				uint8_t off = 0;

				if (ielen != 22)
					return -1;

				ht_op = p + 1;
				off = ht_op[0] & 0x3;

				if (off == 0)
					*bw = BW20;
				else if (off == 1 || off == 3)
					*bw = BW40;

				if ((ht_op[0] & bit(2)) == 0)
					*bw = BW20;
				cntr++;
			}
			break;
		case 192:  /* VHT oper */
			{
				uint8_t *vht_op;
				uint8_t w, cfs0, cfs1;

				if (ielen < 5)
					return -1;

				vht_op = p;
				w = vht_op[0];
				cfs0 = vht_op[1];
				cfs1 = vht_op[2];

				if (w == 0)
					break;

				if (cfs1 == 0) {
					*bw = BW80;
				} else if (cfs1 > 0){
					if (cfs1 - cfs0 == 8)
						*bw = BW160;
					else if (cfs1 - cfs0 == 16)
						*bw = BW8080;
				}
				cntr++;
			}
			break;
		default:
			break;
		}

		if (cntr == 2)
			break;

		offset += ies[offset + 1] + 2;
	}
	return 0;
}

static int _nlwifi_cmdresp(struct nl_msg *msg, void *resp)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct cmd_respdata {
		uint32_t attr;
		int len;
		void *data;
	} *res = resp;

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_IFNAME])
		return NL_SKIP;

	if (!tb[res->attr])
		return NL_SKIP;

	/* libwifi_dbg("res->attr = %u\n", res->attr); */
	switch (res->attr) {
	case NL80211_ATTR_CHANNEL_WIDTH:
		{
			uint32_t bw;

			bw = nla_get_u32(tb[NL80211_ATTR_CHANNEL_WIDTH]);
			*((uint32_t *)res->data) = bw;
		}
		break;
	case NL80211_ATTR_WIPHY_FREQ:
		{
			uint32_t freq;

			freq = nla_get_u32(tb[NL80211_ATTR_WIPHY_FREQ]);
			*((uint32_t *)res->data) = freq;
		}
		break;
	case NL80211_ATTR_IFTYPE:
		{
			uint32_t iftype;

			iftype = nla_get_u32(tb[NL80211_ATTR_IFTYPE]);
			*((uint32_t *)res->data) = iftype;
		}
		break;
	default:
		/* libwifi_dbg("res->attr = default\n"); */
		res->len = nla_len(tb[res->attr]);
		memcpy(res->data, nla_data(tb[res->attr]), res->len);
		break;
	}

	return NL_SKIP;
}

int nlwifi_get_ssid(const char *ifname, char *ssid)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	struct cmd_respdata {
		uint32_t attr;
		int len;
		void *data;
	} resp = {
		.attr = NL80211_ATTR_SSID,
		.len = 0,
		.data = ssid,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!ssid)
		return -EINVAL;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_INTERFACE, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_cmdresp, &resp);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
	/* libwifi_dbg("SSID = %s\n", ssid); */

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_get_bssid(const char *ifname, uint8_t *bssid)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	uint8_t macaddr[6] = {0};
	struct cmd_respdata {
		uint32_t attr;
		int len;
		void *data;
	} res = {
		.attr = NL80211_ATTR_MAC,
		.len = 0,
		.data = macaddr,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!bssid)
		return -EINVAL;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_INTERFACE, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_cmdresp, &res);
	if (ret < 0 || res.len == 0) {
		ret = -1;
		goto nla_put_failure;
	}

	ret = 0;
	memcpy(bssid, macaddr, 6);
	/* snprintf(bssid, 18, "%02X:%02X:%02X:%02X:%02X:%02X\n",
			macaddr[0], macaddr[1], macaddr[2],
			macaddr[3], macaddr[4], macaddr[5]); */

	/* libwifi_dbg("BSSID = %s\n", bssid); */

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;

}

int nlwifi_get_chan(const char *ifname, uint32_t *channel)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	uint32_t freq = 0;
	struct cmd_respdata {
		uint32_t attr;
		int len;
		void *data;
	} resp = {
		.attr = NL80211_ATTR_WIPHY_FREQ,
		.len = sizeof(freq),
		.data = &freq,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!channel)
		return -EINVAL;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_INTERFACE, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_cmdresp, &resp);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;

	*channel = ieee80211_frequency_to_channel(freq);
	libwifi_dbg("freq = %u chan %u\n", freq, *channel);

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

static int _nlwifi_get_channels(struct nl_msg *msg, void *resp)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct nlattr *tb_band[NL80211_BAND_ATTR_MAX + 1];
	struct nlattr *tb_freq[NL80211_FREQUENCY_ATTR_MAX + 1];
	struct nlattr *nl_band, *nl_freq;
	struct cmd_respdata {
		struct {
			int band;
			int bw;
		} req;
		int nr;
		void *data;
	} *res = resp;
	int iter_band;
	int iter_freq;

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_WIPHY_BANDS])
		return NL_SKIP;

	nla_for_each_nested(nl_band, tb[NL80211_ATTR_WIPHY_BANDS], iter_band) {
		nla_parse(tb_band,
			NL80211_BAND_ATTR_MAX,
			nla_data(nl_band),
			nla_len(nl_band),
			NULL);

		if (!tb_band[NL80211_BAND_ATTR_FREQS])
			continue;

		nla_for_each_nested(nl_freq, tb_band[NL80211_BAND_ATTR_FREQS],
							iter_freq) {
			uint32_t freq;
			int ch;

			nla_parse(tb_freq,
				NL80211_FREQUENCY_ATTR_MAX,
				nla_data(nl_freq),
				nla_len(nl_freq),
				NULL);

			if (!tb_freq[NL80211_FREQUENCY_ATTR_FREQ])
				continue;

			if (tb_freq[NL80211_FREQUENCY_ATTR_DISABLED])
				continue;

			freq = nla_get_u32(tb_freq[NL80211_FREQUENCY_ATTR_FREQ]);
			ch = wifi_freq_to_channel(freq);
			libwifi_dbg("Freq: %d MHz  (Ch = %d)\n", freq, ch);
			((uint32_t *)res->data)[res->nr] = ch;
			res->nr++;
		}
	}

	return NL_SKIP;
}

int nlwifi_get_supp_channels(const char *ifname, uint32_t *chlist, int *num,
					const char *cc,
					enum wifi_band band,
					enum wifi_bw bw)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	uint32_t ch[128] = {0};
	int devidx;
	int ret = -1;
	struct cmd_respdata {
		struct {
			int band;
			int bw;
		} req;
		int nr;
		void *data;
	} res = {
		//.attr = NL80211_FREQUENCY_ATTR_FREQ, /* NL80211_ATTR_UNSPEC */
		.req.band = band,
		.req.bw = bw,
		.nr = 0,
		.data = ch,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_WIPHY, 0, 0);
	if (!msg)
		goto out_msg_failure;

	/* prepare cmd */
	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_get_channels, &res);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
	*num = res.nr;
	memcpy(chlist, res.data, res.nr * sizeof(int));

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_get_bandwidth(const char *ifname, enum wifi_bw *bandwidth)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	uint32_t bw = 0;
	struct cmd_respdata {
		uint32_t attr;
		int len;
		void *data;
	} res = {
		.attr = NL80211_ATTR_CHANNEL_WIDTH,
		.len = sizeof(bw),
		.data = &bw,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_INTERFACE, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, _nlwifi_cmdresp, &res);
	if (ret < 0 || res.len == 0) {
		ret = -1;
		goto nla_put_failure;
	}

	ret = 0;

	if (bw == NL80211_CHAN_WIDTH_80)
		*bandwidth = BW80;
	else if (bw == NL80211_CHAN_WIDTH_40)
		*bandwidth = BW40;
	else if (bw == NL80211_CHAN_WIDTH_20 ||
			bw == NL80211_CHAN_WIDTH_20_NOHT)
		*bandwidth = BW20;
	else if (bw == NL80211_CHAN_WIDTH_160)
		*bandwidth = BW160;
	else if (bw == NL80211_CHAN_WIDTH_80P80)
		*bandwidth = BW8080;
	else
		*bandwidth = BW_UNKNOWN;

	libwifi_dbg("nlwifi bandwidth = %d\n", *bandwidth);

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_get_channel(const char *ifname, uint32_t *channel,
		       enum wifi_bw *bw)
{
	int ret;

	ret = nlwifi_get_chan(ifname, channel);
	if (ret) {
		libwifi_err("nlwifi_get_chan(%s) failed %d\n", ifname, ret);
		return ret;
	}

	return nlwifi_get_bandwidth(ifname, bw);
}

int nlwifi_get_mode(const char *ifname, enum wifi_mode *mode)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	uint32_t nl_iftype = 0;
	struct cmd_respdata {
		uint32_t attr;
		int len;
		void *data;
	} resp = {
		.attr = NL80211_ATTR_IFTYPE,
		.len = sizeof(nl_iftype),
		.data = &nl_iftype,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!mode)
		return -EINVAL;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_INTERFACE, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_cmdresp, &resp);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;

	*mode = ieee80211_type_to_wifi_mode((enum nl80211_iftype) nl_iftype);
	libwifi_dbg("nl_type=%d wifi_mode=%d\n", nl_iftype, *mode);

nla_put_failure:
	nlmsg_free(msg);

out_msg_failure:
	nl_socket_free(nl);

out:
	return ret;
}

static int _nlwifi_vendor_cmdresp(struct nl_msg *msg, void *resp)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct nlattr *nl_resp, *nl_iter;
	struct cmd_respdata {
		int *len;
		void *data;
	} *res = resp;
	int iter = 0;

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	*res->len = 0;
	nl_resp = tb[NL80211_ATTR_VENDOR_DATA];
	if (!nl_resp)
		return NL_SKIP;

	nla_for_each_nested(nl_iter, nl_resp, iter) {
		memcpy(res->data + *res->len, nla_data(nl_iter),
						nla_len(nl_iter));
		*res->len += nla_len(nl_iter);
	}

	return NL_SKIP;
}

int nlwifi_vendor_cmd(const char *ifname, uint32_t vid, uint32_t subcmd,
					uint8_t *in, int ilen,
					uint8_t *out, int *olen)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	struct cmd_respdata {
		int *len;
		void *data;
	} resp = {
		.len = olen,
		.data = out,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_VENDOR, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	NLA_PUT_U32(msg, NL80211_ATTR_VENDOR_ID, vid);
	NLA_PUT_U32(msg, NL80211_ATTR_VENDOR_SUBCMD, subcmd);
	if (in)
		NLA_PUT(msg, NL80211_ATTR_VENDOR_DATA, ilen, in);

	ret = nlwifi_send_msg(nl, msg, _nlwifi_vendor_cmdresp, &resp);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

uint32_t nlwifi_bw_enum2MHz(enum wifi_bw bw)
{
	uint32_t bandwidth = 0;
	switch(bw) {
	case BW80:
		bandwidth = 80;
		break;
	case BW40:
		bandwidth = 40;
		break;
	case BW20:
		bandwidth = 20;
		break;
	case BW160:
	case BW8080:
		bandwidth = 160;
		break;
	default:
		libwifi_warn("unknown bandwidth %u\n", bw);
		break;
	}

	return bandwidth;
}

int nlwifi_radio_get_maxrate(const char *ifname, unsigned long *rate_Mbps)
{
	struct wifi_radio radio;
	int ret = 0;
	int max_mcs = 0;

	ret = nlwifi_radio_info(ifname, &radio);
	if (ret != 0)
		goto out;

	if (radio.oper_band == BAND_2) max_mcs = 7;
	if (radio.oper_band == BAND_5) max_mcs = 9;

	*rate_Mbps = wifi_mcs2rate(max_mcs, nlwifi_bw_enum2MHz(radio.curr_bw), radio.rx_streams, 1);

out:
	return ret;
}

int nlwifi_ap_get_caps(const char *ifname, struct wifi_caps *caps)
{
	//TODO
	return 0;
}

static int _nlwifi_radio_get_caps(struct nl_msg *msg, void *resp)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct nlattr *tb_band[NL80211_BAND_ATTR_MAX + 1];
	struct nlattr *nl_band;
	struct wifi_caps *caps = resp;
	int iter_band;

#define min(x, y)	(x) < (y) ? (x) : (y)

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_WIPHY_BANDS])
		return NL_SKIP;

	if (tb[NL80211_ATTR_EXT_CAPA]) {
		caps->valid |= WIFI_CAP_EXT_VALID;
		memcpy(caps->ext.byte,
				nla_data(tb[NL80211_ATTR_EXT_CAPA]),
				min(nla_len(tb[NL80211_ATTR_EXT_CAPA]), 16));
	}

	nla_for_each_nested(nl_band, tb[NL80211_ATTR_WIPHY_BANDS], iter_band) {
		nla_parse(tb_band,
			NL80211_BAND_ATTR_MAX,
			nla_data(nl_band),
			nla_len(nl_band),
			NULL);

		if (tb_band[NL80211_BAND_ATTR_HT_CAPA]) {
			caps->valid |= WIFI_CAP_HT_VALID;
			memcpy(caps->ht.byte,
				nla_data(tb_band[NL80211_BAND_ATTR_HT_CAPA]),
				nla_len(tb_band[NL80211_BAND_ATTR_HT_CAPA]));

			if (caps->ht.cap == ntohs(caps->ht.cap))
				caps->ht.cap = bswap_16(caps->ht.cap);
		}
		if (tb_band[NL80211_BAND_ATTR_VHT_CAPA]) {
			caps->valid |= WIFI_CAP_VHT_VALID;
			memcpy(caps->vht.byte,
				nla_data(tb_band[NL80211_BAND_ATTR_VHT_CAPA]),
				nla_len(tb_band[NL80211_BAND_ATTR_VHT_CAPA]));

			if (caps->vht.cap == ntohl(caps->vht.cap))
				caps->vht.cap = bswap_32(caps->vht.cap);
		}
	}

	//TODO:
	// NL80211_ATTR_FEATURE_FLAGS
	// NL80211_ATTR_EXT_FEATURES,
	// NL80211_ATTR_IFTYPE_EXT_CAPA (per-AP caps can override radio caps)

	return NL_SKIP;
}

int nlwifi_radio_get_caps(const char *ifname, struct wifi_caps *caps)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_WIPHY, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, _nlwifi_radio_get_caps, caps);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

static int _nlwifi_radio_info(struct nl_msg *msg, void *resp)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct wifi_radio *radio = resp;

#define min(x, y)	(x) < (y) ? (x) : (y)

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_WIPHY_BANDS])
		return NL_SKIP;

	if (tb[NL80211_ATTR_WIPHY_FRAG_THRESHOLD])
		radio->frag = nla_get_u32(tb[NL80211_ATTR_WIPHY_FRAG_THRESHOLD]);

	if (tb[NL80211_ATTR_WIPHY_RTS_THRESHOLD])
		radio->rts = nla_get_u32(tb[NL80211_ATTR_WIPHY_RTS_THRESHOLD]);

	if (tb[NL80211_ATTR_WIPHY_RETRY_SHORT])
		radio->srl = nla_get_u8(tb[NL80211_ATTR_WIPHY_RETRY_SHORT]);

	if (tb[NL80211_ATTR_WIPHY_RETRY_LONG])
		radio->lrl = nla_get_u8(tb[NL80211_ATTR_WIPHY_RETRY_LONG]);

	if (tb[NL80211_ATTR_BEACON_INTERVAL])
		radio->beacon_int = nla_get_u32(tb[NL80211_ATTR_BEACON_INTERVAL]);

	if (tb[NL80211_ATTR_DTIM_PERIOD])
		radio->dtim_period = nla_get_u32(tb[NL80211_ATTR_DTIM_PERIOD]);

	if (tb[NL80211_ATTR_WIPHY_ANTENNA_TX]) {
		uint32_t tmp;
		int i;

		tmp = nla_get_u32(tb[NL80211_ATTR_WIPHY_ANTENNA_TX]);
		for (i = 0; i < 8; i++) {
			if (!!(tmp & (1 << i)))
				radio->tx_streams++;
		}
	}

	if (tb[NL80211_ATTR_WIPHY_ANTENNA_RX]) {
		uint32_t tmp;
		int i;

		tmp = nla_get_u32(tb[NL80211_ATTR_WIPHY_ANTENNA_RX]);
		for (i = 0; i < 8; i++) {
			if (!!(tmp & (1 << i)))
				radio->rx_streams++;
		}
	}

	return NL_SKIP;
}

int nlwifi_radio_info(const char *name, struct wifi_radio *radio)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	uint32_t curr_ch = 0;
	uint32_t chs[64] = {0};
	int devidx;
	int num_chs = 64;
	int ret = -1;
	int i;

	devidx = if_nametoindex(name);
	if (devidx == 0) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_WIPHY, 0, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, _nlwifi_radio_info, radio);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
	ret |= nlwifi_get_channel(name, &curr_ch, &radio->curr_bw);
	ret |= nlwifi_get_bandwidth(name, &radio->curr_bw);
	if (!ret) {
		radio->channel = curr_ch;
		radio->oper_band = radio->channel > 0 && radio->channel <= 14 ?
			BAND_2 : radio->channel >= 36 && radio->channel < 200 ?
			BAND_5 : BAND_UNKNOWN;

		nlwifi_get_country(name, radio->regdomain);
		ret = nlwifi_get_supp_channels(name, chs, &num_chs,
				radio->regdomain, radio->oper_band,
				radio->curr_bw);
		if (!ret) {
			for (i = 0; i < num_chs; i++)
				radio->supp_channels[i] = chs[i];
		}
	}

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

static int _nlwifi_get_assoclist(struct nl_msg *msg, void *assoclist)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct asslist {
		const char *vif;
		int i;
		int nr;
		uint8_t *macs;
	} *stalist = assoclist;
	uint8_t *e;
	char ifname[16] = {0};

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);


	if (stalist->i > stalist->nr) {
		libwifi_warn("Num stations > %d !\n", stalist->nr);
		return NL_SKIP;
	}


	if_indextoname(nla_get_u32(tb[NL80211_ATTR_IFINDEX]), ifname);
	if (!tb[NL80211_ATTR_MAC])
		return NL_SKIP;

	if (!strncmp(stalist->vif, ifname, 15)) {
		const uint8_t *mac = nla_data(tb[NL80211_ATTR_MAC]);

		e = stalist->macs + stalist->i * 6;
		memcpy(e, mac, 6);
		stalist->i += 1;
	}

	return NL_SKIP;
}

int nlwifi_get_assoclist(const char *ifname, uint8_t *stas, int *num_stas)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	struct asslist {
		const char *vif;
		int i;
		int nr;
		uint8_t *macs;
	} stalist = {
		.vif = ifname,
		.i = 0,
		.nr = *num_stas,
		.macs = stas,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!stas || *num_stas <= 0) {
		libwifi_err("Invalid args!\n");
		return -EINVAL;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_STATION, NLM_F_DUMP, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, _nlwifi_get_assoclist, &stalist);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
	*num_stas = stalist.i;

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

static int nlwifi_bitrate_attrs(struct nlattr *attr, struct wifi_rate *rate)
{
	struct nlattr *r[NL80211_RATE_INFO_MAX + 1];
	static struct nla_policy bitrate_policy[NL80211_RATE_INFO_MAX + 1] = {
		[NL80211_RATE_INFO_BITRATE] = { .type = NLA_U16 },
		[NL80211_RATE_INFO_MCS] = { .type = NLA_U8 },
		[NL80211_RATE_INFO_40_MHZ_WIDTH] = { .type = NLA_FLAG },
		[NL80211_RATE_INFO_SHORT_GI] = { .type = NLA_FLAG },
		[NL80211_RATE_INFO_BITRATE32] = { .type = NLA_U32 },
		[NL80211_RATE_INFO_VHT_MCS] = { .type = NLA_U8 },
		[NL80211_RATE_INFO_VHT_NSS] = { .type = NLA_U8 },
		[NL80211_RATE_INFO_80_MHZ_WIDTH] = { .type = NLA_FLAG },
		[NL80211_RATE_INFO_80P80_MHZ_WIDTH] = { .type = NLA_FLAG },
		[NL80211_RATE_INFO_160_MHZ_WIDTH] = { .type = NLA_FLAG },
	};
	uint32_t r100kbps = 0;
	int ret = 0;

	ret = nla_parse_nested(r, NL80211_RATE_INFO_MAX, attr, bitrate_policy);
	if (ret) {
		libwifi_err("Error parsing nested bitrate attr %s\n", nl_geterror(ret));
		return ret;
	}

	if (r[NL80211_RATE_INFO_BITRATE32])
		r100kbps= nla_get_u32(r[NL80211_RATE_INFO_BITRATE32]);
	else if (r[NL80211_RATE_INFO_BITRATE])
		r100kbps = nla_get_u16(r[NL80211_RATE_INFO_BITRATE]);

	rate->rate = r100kbps / 10;

	if (r[NL80211_RATE_INFO_SHORT_GI])
		rate->m.sgi = 1;

	if (r[NL80211_RATE_INFO_MCS]) {
		rate->m.mcs = nla_get_u8(r[NL80211_RATE_INFO_MCS]);
		rate->phy = PHY_HT;
	}

	if (r[NL80211_RATE_INFO_VHT_MCS]) {
		rate->m.mcs = nla_get_u8(r[NL80211_RATE_INFO_VHT_MCS]);
		rate->phy = PHY_VHT;
	}

	if (r[NL80211_RATE_INFO_VHT_NSS])
		rate->m.nss = nla_get_u8(r[NL80211_RATE_INFO_VHT_NSS]);

	if (r[NL80211_RATE_INFO_40_MHZ_WIDTH])
		rate->m.bw = 40;
	else if (r[NL80211_RATE_INFO_80_MHZ_WIDTH])
		rate->m.bw = 80;
	else if (r[NL80211_RATE_INFO_160_MHZ_WIDTH])
		rate->m.bw = 160;
	else if (r[NL80211_RATE_INFO_80P80_MHZ_WIDTH])
		rate->m.bw = 160;
	else
		rate->m.bw = 0;

	return 0;
}

static int nlwifi_bss_param_attr(struct nlattr *bss_param_attr, struct wifi_sta *sta)
{
	struct nlattr *bss_info[NL80211_STA_BSS_PARAM_MAX + 1];
	static struct nla_policy bss_poilcy[NL80211_STA_BSS_PARAM_MAX + 1] = {
		[NL80211_STA_BSS_PARAM_SHORT_PREAMBLE] = { .type = NLA_FLAG },
		[NL80211_STA_BSS_PARAM_SHORT_SLOT_TIME] = { .type = NLA_FLAG },
	};

	if (nla_parse_nested(bss_info, NL80211_STA_BSS_PARAM_MAX,
		bss_param_attr, bss_poilcy)) {
		return -1;
	}

	if (bss_info[NL80211_STA_BSS_PARAM_SHORT_SLOT_TIME]) {
		if (nla_get_u16(bss_info[NL80211_STA_BSS_PARAM_SHORT_SLOT_TIME])) {
			wifi_cap_set(sta->cbitmap, WIFI_CAP_SHORT_SLOT);
		}
	}

	return 0;
}


static int cmd_get_station_parse(struct nl_msg *msg, void *info)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct stainfo {
		const char *vif;
		const uint8_t *macaddr;
		struct wifi_sta *s;
	} *resp = info;
	struct wifi_sta *sta = resp->s;
	struct nlattr *s[NL80211_STA_INFO_MAX + 1];

	static struct nla_policy sta_policy[NL80211_STA_INFO_MAX + 1] = {
		[NL80211_STA_INFO_INACTIVE_TIME] = { .type = NLA_U32 },
		[NL80211_STA_INFO_RX_BYTES] = { .type = NLA_U32 },
		[NL80211_STA_INFO_TX_BYTES] = { .type = NLA_U32 },
		[NL80211_STA_INFO_SIGNAL] = { .type = NLA_U8 },
		[NL80211_STA_INFO_TX_BITRATE] = { .type = NLA_NESTED },
		[NL80211_STA_INFO_RX_PACKETS] = { .type = NLA_U32 },
		[NL80211_STA_INFO_TX_PACKETS] = { .type = NLA_U32 },
		[NL80211_STA_INFO_TX_RETRIES] = { .type = NLA_U32 },
		[NL80211_STA_INFO_TX_FAILED] = { .type = NLA_U32 },
		[NL80211_STA_INFO_SIGNAL_AVG] = { .type = NLA_U8 },
		[NL80211_STA_INFO_RX_BITRATE] = { .type = NLA_NESTED },
		[NL80211_STA_INFO_BSS_PARAM] = { .type = NLA_NESTED },
		[NL80211_STA_INFO_CONNECTED_TIME] = { .type = NLA_U32},
		[NL80211_STA_INFO_STA_FLAGS] =
			{ .minlen = sizeof(struct nl80211_sta_flag_update) },
		[NL80211_STA_INFO_RX_BYTES64] = { .type = NLA_U64 },
		[NL80211_STA_INFO_TX_BYTES64] = { .type = NLA_U64 },
		[NL80211_STA_INFO_CHAIN_SIGNAL] = { .type = NLA_NESTED },
		[NL80211_STA_INFO_EXPECTED_THROUGHPUT] = { .type = NLA_U32 },
		[NL80211_STA_INFO_RX_DROP_MISC] = { .type = NLA_U64},
	};
	const uint8_t *mac;
	char ifname[16] = {0};
	struct nl80211_sta_flag_update *sta_flags;

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if_indextoname(nla_get_u32(tb[NL80211_ATTR_IFINDEX]), ifname);
	if (!tb[NL80211_ATTR_MAC] || !tb[NL80211_ATTR_STA_INFO])
		return NL_SKIP;

	if (strncmp(resp->vif, ifname, 15))
		return NL_SKIP;    /* not ours */

	mac = nla_data(tb[NL80211_ATTR_MAC]);
	if (resp->macaddr && memcmp(mac, resp->macaddr, 6))
		return NL_SKIP;
	memcpy(sta->macaddr, mac, 6);

	if (nla_parse_nested(s, NL80211_STA_INFO_MAX, tb[NL80211_ATTR_STA_INFO],
				sta_policy)) {
		libwifi_warn("Error parsing NL80211_STA_INFO attrs!\n");
		return NL_SKIP;
	}

	memcpy(sta->bssid, nla_data(tb[NL80211_ATTR_MAC]), 6);

	if (s[NL80211_STA_INFO_INACTIVE_TIME])
		sta->idle_time =
			nla_get_u32(s[NL80211_STA_INFO_INACTIVE_TIME]) / 1000;

	if (s[NL80211_STA_INFO_CONNECTED_TIME])
		sta->conn_time =
			nla_get_u32(s[NL80211_STA_INFO_CONNECTED_TIME]);

	if (s[NL80211_STA_INFO_RX_BITRATE])
		nlwifi_bitrate_attrs(s[NL80211_STA_INFO_RX_BITRATE], &sta->rx_rate);

	if (s[NL80211_STA_INFO_TX_BITRATE])
		nlwifi_bitrate_attrs(s[NL80211_STA_INFO_TX_BITRATE], &sta->tx_rate);

	if (s[NL80211_STA_INFO_RX_BYTES64]) {
		sta->stats.rx_bytes =
			nla_get_u64(s[NL80211_STA_INFO_RX_BYTES64]);
	} else if (s[NL80211_STA_INFO_RX_BYTES]) {
		sta->stats.rx_bytes =
			nla_get_u32(s[NL80211_STA_INFO_RX_BYTES]);
	}

	if (s[NL80211_STA_INFO_TX_BYTES64]) {
		sta->stats.tx_bytes =
			nla_get_u64(s[NL80211_STA_INFO_TX_BYTES64]);
	} else if (s[NL80211_STA_INFO_TX_BYTES]) {
		sta->stats.tx_bytes =
			nla_get_u32(s[NL80211_STA_INFO_TX_BYTES]);
	}

	if (s[NL80211_STA_INFO_RX_PACKETS])
		sta->stats.rx_pkts =
			nla_get_u32(s[NL80211_STA_INFO_RX_PACKETS]);

	if (s[NL80211_STA_INFO_TX_PACKETS])
		sta->stats.tx_pkts =
			nla_get_u32(s[NL80211_STA_INFO_TX_PACKETS]);

	if (s[NL80211_STA_INFO_TX_RETRIES])
		sta->stats.tx_retry_pkts =
			nla_get_u32(s[NL80211_STA_INFO_TX_RETRIES]);

	if (s[NL80211_STA_INFO_TX_FAILED])
		sta->stats.tx_fail_pkts =
			nla_get_u32(s[NL80211_STA_INFO_TX_FAILED]);

	if (s[NL80211_STA_INFO_RX_DROP_MISC])
		sta->stats.rx_fail_pkts =
			nla_get_u64(s[NL80211_STA_INFO_RX_DROP_MISC]);


	if (s[NL80211_STA_INFO_SIGNAL]) {
		sta->rssi[0] = nla_get_u8(s[NL80211_STA_INFO_SIGNAL]);
	}

	if (s[NL80211_STA_INFO_CHAIN_SIGNAL]) {
		struct nlattr *itr_attr;
		struct nlattr *rssi_attrs = s[NL80211_STA_INFO_CHAIN_SIGNAL];
		int idx = 0;
		int rem;

		nla_for_each_nested(itr_attr, rssi_attrs, rem) {
			sta->rssi[idx++] = nla_get_u8(itr_attr);
			if (idx == WIFI_NUM_ANTENNA)
				break;
		}
	}

	if (s[NL80211_STA_INFO_SIGNAL_AVG])
		sta->rssi_avg = (int)nla_get_u8(s[NL80211_STA_INFO_SIGNAL_AVG]);

	if (s[NL80211_STA_INFO_STA_FLAGS]) {
		sta_flags = (struct nl80211_sta_flag_update *)
					nla_data(s[NL80211_STA_INFO_STA_FLAGS]);

		if (sta_flags->mask & BIT(NL80211_STA_FLAG_SHORT_PREAMBLE)) {
			if (sta_flags->set & BIT(NL80211_STA_FLAG_SHORT_PREAMBLE)) {
				wifi_cap_set(sta->cbitmap, WIFI_CAP_SHORT_PREAMBLE);
			}
		}
		if (sta_flags->mask & BIT(NL80211_STA_FLAG_WME)) {
			wifi_cap_set(sta->cbitmap, WIFI_CAP_WMM);
			if (sta_flags->set & BIT(NL80211_STA_FLAG_WME)) {
				wifi_status_set(sta->sbitmap, WIFI_STATUS_WMM);
			}
		}
	}

	if (s[NL80211_STA_INFO_BSS_PARAM]) {
		nlwifi_bss_param_attr(s[NL80211_STA_INFO_BSS_PARAM], sta);
	}

	if (s[NL80211_STA_INFO_EXPECTED_THROUGHPUT])
		sta->tx_thput =
			nla_get_u32(s[NL80211_STA_INFO_EXPECTED_THROUGHPUT]) / 1024;

	nlwifi_radio_get_maxrate(ifname, (unsigned long *) &sta->maxrate);
	return NL_SKIP;
}

int nlwifi_sta_info(const char *ifname, struct wifi_sta *info)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	struct stainfo {
		const char *vif;
		const uint8_t *macaddr;
		struct wifi_sta *s;
	} sta = {
		.vif = ifname,
		.macaddr = NULL,
		.s = info,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	ret = nlwifi_get_bssid(ifname,info->macaddr);
	if (ret != 0)
		goto out;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_STATION, NLM_F_DUMP, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, cmd_get_station_parse, &sta);
	if (ret < 0)
		goto nla_put_failure;

	ret = nlwifi_radio_get_maxrate(ifname, (unsigned long *) &info->maxrate);
	if (ret < 0)
		goto nla_put_failure;

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_get_sta_info(const char *ifname, uint8_t *addr, struct wifi_sta *info)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	struct stainfo {
		const char *vif;
		const uint8_t *macaddr;
		struct wifi_sta *s;
	} sta = {
		.vif = ifname,
		.macaddr = addr,
		.s = info,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!addr) {
		libwifi_err("Invalid args!\n");
		return -EINVAL;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_STATION, NLM_F_DUMP, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, cmd_get_station_parse, &sta);
	if (ret < 0)
		goto nla_put_failure;

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_sta_get_stats(const char *ifname,struct wifi_sta_stats *s)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;

	struct wifi_sta station = {};
	struct stainfo {
		const char *vif;
		const uint8_t *macaddr;
		struct wifi_sta *s;
	} sta = {
		.vif = ifname,
		.macaddr = NULL,
		.s = &station,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_STATION, NLM_F_DUMP, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);
	ret = nlwifi_send_msg(nl, msg, cmd_get_station_parse, &sta);
	if (ret < 0)
		goto nla_put_failure;

	memcpy(s, &(sta.s->stats), sizeof(struct wifi_sta_stats));

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

static int nlwifi_parse_scan_results(struct nl_msg *msg, void *sres)
{
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *bss[NL80211_BSS_MAX + 1];
	static struct nla_policy bss_policy[NL80211_BSS_MAX + 1] = {
		[NL80211_BSS_BSSID] = { .type = NLA_UNSPEC },
		[NL80211_BSS_FREQUENCY] = { .type = NLA_U32 },
		[NL80211_BSS_TSF] = { .type = NLA_U64 },
		[NL80211_BSS_BEACON_INTERVAL] = { .type = NLA_U16 },
		[NL80211_BSS_CAPABILITY] = { .type = NLA_U16 },
		[NL80211_BSS_INFORMATION_ELEMENTS] = { .type = NLA_UNSPEC },
		[NL80211_BSS_SIGNAL_MBM] = { .type = NLA_U32 },
		[NL80211_BSS_SIGNAL_UNSPEC] = { .type = NLA_U8 },
		[NL80211_BSS_STATUS] = { .type = NLA_U32 },
		[NL80211_BSS_SEEN_MS_AGO] = { .type = NLA_U32 },
		[NL80211_BSS_BEACON_IES] = { .type = NLA_UNSPEC },
	};
	struct scanresult {
		int i;
		int num;
		struct wifi_bss *bsss;
	} *scanres = sres;
	struct wifi_bss *e = scanres->bsss + scanres->i;
	uint8_t *ssid_ie;
	size_t ie_len;
	uint8_t *ie;
	int rssi = 0;
	int ret;
	int i;

	libwifi_dbg("sres: &bsss = %p (numbss limit = %d)\n", e, scanres->num);
	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[NL80211_ATTR_BSS])
		return NL_SKIP;  /* no bssid */

	if (scanres->i > scanres->num) {
		libwifi_warn("Num scan results > %d !\n", scanres->num);
		return NL_SKIP;
	}

	if (nla_parse_nested(bss, NL80211_BSS_MAX, tb[NL80211_ATTR_BSS],
				bss_policy))
		return NL_SKIP;

	if (bss[NL80211_BSS_INFORMATION_ELEMENTS]) {
		ie = nla_data(bss[NL80211_BSS_INFORMATION_ELEMENTS]);
		ie_len = nla_len(bss[NL80211_BSS_INFORMATION_ELEMENTS]);
	} else {
		ie = NULL;
		ie_len = 0;
	}

	if (bss[NL80211_BSS_BSSID]) {
		libwifi_dbg("BSSID: " MACFMT "  ",
			MAC2STR((uint8_t *)nla_data(bss[NL80211_BSS_BSSID])));
		e = scanres->bsss + scanres->i;
		memset(e, 0, sizeof(struct wifi_bss));
		memcpy(e->bssid, nla_data(bss[NL80211_BSS_BSSID]), 6);
	}

	if (bss[NL80211_BSS_FREQUENCY]) {
		uint32_t frequency;
		libwifi_dbg("Freq: %u Mhz      ",
				nla_get_u32(bss[NL80211_BSS_FREQUENCY]));
		frequency = nla_get_u32(bss[NL80211_BSS_FREQUENCY]);
		e->channel = wifi_freq_to_channel(frequency);
		if (e->channel > 0 && e->channel <= 14)
			e->band = BAND_2;
		else if (e->channel >= 36 && e->channel < 200)
			e->band = BAND_5;
		else
			e->band = BAND_UNKNOWN;
	}

	if (bss[NL80211_BSS_BEACON_INTERVAL]) {
		e->beacon_int = nla_get_u16(bss[NL80211_BSS_BEACON_INTERVAL]);
		libwifi_dbg("    Bcn Int: %d  ",
				nla_get_u16(bss[NL80211_BSS_BEACON_INTERVAL]));
	}

	if (bss[NL80211_BSS_CAPABILITY]) {
		e->caps.basic.cap = nla_get_u16(bss[NL80211_BSS_CAPABILITY]);
		e->caps.valid |= WIFI_CAP_BASIC_VALID;
	}

	if (bss[NL80211_BSS_SIGNAL_MBM]) {
		rssi = nla_get_u32(bss[NL80211_BSS_SIGNAL_MBM]);
		rssi /= 100;	/* in dBm */
	} else if (bss[NL80211_BSS_SIGNAL_UNSPEC]) {
		rssi = nla_get_u8(bss[NL80211_BSS_SIGNAL_UNSPEC]);
	}
	e->rssi = rssi;
	libwifi_dbg("    Rssi: %d dBm", rssi);

	ssid_ie = wifi_find_ie(ie, ie_len, 0);
	if (ssid_ie) {
		libwifi_dbg("     Ssid: ");
		for (i = 0; i < ssid_ie[1]; i++)
			libwifi_dbg("%c", ssid_ie[2+i]);

		memcpy(e->ssid, &ssid_ie[2], ssid_ie[1]);
	}

	wifi_get_bss_security_from_ies(e, ie, ie_len);
	libwifi_dbg("       wpa_versions 0x%x pairwise = 0x%x, group = 0x%x akms 0x%x caps 0x%x\n",
		    e->rsn.wpa_versions, e->rsn.pair_ciphers, e->rsn.group_cipher,
		    e->rsn.akms, e->rsn.rsn_caps);

	ret = nlwifi_get_ie(ie, ie_len, 11, &e->load);

	ret = nlwifi_get_ie(ie, ie_len, 127, &e->caps.ext);
	if (!ret)
		e->caps.valid |= WIFI_CAP_EXT_VALID;

	ret = nlwifi_get_ie(ie, ie_len, 45, &e->caps.ht);
	if (!ret)
		e->caps.valid |= WIFI_CAP_HT_VALID;

	ret = nlwifi_get_ie(ie, ie_len, 191, &e->caps.vht);
	if (!ret)
		e->caps.valid |= WIFI_CAP_VHT_VALID;

	ret = nlwifi_get_ie(ie, ie_len, 70, &e->caps.rrm);
	if (!ret)
		e->caps.valid |= WIFI_CAP_RM_VALID;

	nlwifi_get_bandwidth_from_ie(ie, ie_len, &e->curr_bw);

	//TODO: wifi supp standards
	//TODO: a-only, b-only, g-only etc.
	if (!!(e->caps.valid & WIFI_CAP_VHT_VALID))
		e->oper_std = WIFI_N | WIFI_AC;
	else if (!!(e->caps.valid & WIFI_CAP_HT_VALID))
		e->oper_std = e->channel > 14 ? (WIFI_A | WIFI_N) :
						(WIFI_B | WIFI_G | WIFI_N);
	else
		e->oper_std = e->channel > 14 ? WIFI_A : (WIFI_B | WIFI_G);

	scanres->i += 1;

	return NL_SKIP;
}

int nlwifi_get_scan_results(const char *ifname, struct wifi_bss *bsss,
						int *num)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int devidx;
	int ret = -1;
	struct scanres {
		int i;
		int nr;
		struct wifi_bss *bsslist;
	} scanres = {
		.i = 0,
		.nr = *num,
		.bsslist = bsss,
	};

	devidx = if_nametoindex(ifname);
	if (!devidx) {
		ret = -errno;
		goto out;
	}

	if (!bsss || *num <= 0)
		goto out;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_SCAN, NLM_F_DUMP, 0);
	if (!msg)
		goto out_msg_failure;

	NLA_PUT_U32(msg, NL80211_ATTR_IFINDEX, devidx);

	ret = nlwifi_send_msg(nl, msg, nlwifi_parse_scan_results, &scanres);
	if (ret < 0)
		goto nla_put_failure;

	ret = 0;
	*num = scanres.i;
	libwifi_dbg("Num scanned bsss = %d   bsss = %p\n", *num, bsss);

nla_put_failure:
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}


static int _nlwifi_get_country(struct nl_msg *msg, void *resp)
{
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *tb[NL80211_ATTR_MAX + 1];

	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);

	if (tb[NL80211_ATTR_REG_ALPHA2])
		memcpy(resp, nla_data(tb[NL80211_ATTR_REG_ALPHA2]),
				nla_len(tb[NL80211_ATTR_REG_ALPHA2]));

	return NL_SKIP;
}

int nlwifi_get_country(const char *name, char *alpha2)
{
	struct nl_sock *nl;
	struct nl_msg *msg;
	int ret = -1;

	nl = nlwifi_socket();
	if (nl <= 0)
		goto out;

	msg = nlwifi_alloc_msg(nl, NL80211_CMD_GET_REG, 0, 0);
	if (!msg)
		goto out_msg_failure;

	ret = nlwifi_send_msg(nl, msg, _nlwifi_get_country, alpha2);
	if (ret < 0)
		goto out_msg_failure;

	ret = 0;
	nlmsg_free(msg);
out_msg_failure:
	nl_socket_free(nl);
out:
	return ret;
}

int nlwifi_get_noise(const char *name, int *noise)
{
	/* TODO Get this from survey data */
	return -1;
}

int nlwifi_get_supp_stds(const char *name, uint8_t *std)
{
	struct wifi_radio radio;
	struct wifi_caps caps;

	if (WARN_ON(nlwifi_radio_info(name, &radio)))
		return -1;
	if (WARN_ON(nlwifi_radio_get_caps(name, &caps)))
		return -1;

	switch (radio.oper_band) {
	case BAND_2:
		*std |= WIFI_G | WIFI_B;
		if (caps.valid & WIFI_CAP_HT_VALID)
			*std |= WIFI_N;
		break;
	case BAND_5:
		*std |= WIFI_A;
		if (caps.valid & WIFI_CAP_HT_VALID)
			*std |= WIFI_N;
		if (caps.valid & WIFI_CAP_VHT_VALID)
			*std |= WIFI_AC;
		break;
	default:
		libwifi_warn("[%s] %s oper_band %u\n", name, __func__, radio.oper_band);
		break;
	}

	return 0;
}

/* events handling */
struct nlwifi_event {
	struct nl_sock *sock;
	struct event_struct req;  /* as passed by caller */
};

int nlwifi_event_handler(struct nl_msg *msg, void *arg)
{
	struct nlwifi_event *wevt = (struct nlwifi_event *)arg;
	struct event_response *resp = &wevt->req.resp;
	struct nlmsghdr *nlh = nlmsg_hdr(msg);
	struct genlmsghdr *gnlh = nlmsg_data(nlh);
	struct nlattr *nlattrs[NUM_NL80211_ATTR];
	char ifname[16] = {0};
	int ret;

	if (!genlmsg_valid_hdr(nlh, 0)) {
		libwifi_err("received invalid message\n");
		return 0;
	}

	ret = nla_parse(nlattrs, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
			genlmsg_attrlen(gnlh, 0), NULL);
	if (ret)
		return ret;

	if_indextoname(nla_get_u32(nlattrs[NL80211_ATTR_IFINDEX]), ifname);

	libwifi_dbg("%s %s cmd %u\n", ifname, __func__, gnlh->cmd);

	if (strncmp(ifname, wevt->req.ifname, 16)) {
		/* libwifi_err("Event from '%s' is not ours.\n", ifname); */
		return 0;
	}

	switch (gnlh->cmd) {
	case NL80211_CMD_TRIGGER_SCAN:
		resp->type = WIFI_EVENT_SCAN_START;
		break;
	case NL80211_CMD_NEW_SCAN_RESULTS:
		resp->type = WIFI_EVENT_SCAN_END;
		break;
	case NL80211_CMD_CONNECT:
		resp->type = WIFI_EVENT_STA_ASSOC;
		break;
	case NL80211_CMD_DISCONNECT:
		resp->type = WIFI_EVENT_STA_DISASSOC;
		break;
	case NL80211_CMD_MICHAEL_MIC_FAILURE:
		break;
	case NL80211_CMD_VENDOR:
		{
			struct nlwifi_event_vendor_resp *vr =
				(struct nlwifi_event_vendor_resp *)resp->data;
			uint8_t *data;
			int len;

			resp->type = WIFI_EVENT_VENDOR;
			vr->oui = 0;
			vr->subcmd = 0;

			if (nlattrs[NL80211_ATTR_VENDOR_ID])
				vr->oui = nla_get_u32(nlattrs[NL80211_ATTR_VENDOR_ID]);

			if (nlattrs[NL80211_ATTR_VENDOR_SUBCMD])
				vr->subcmd = nla_get_u32(nlattrs[NL80211_ATTR_VENDOR_SUBCMD]);

			if (nlattrs[NL80211_ATTR_VENDOR_DATA]) {
				data = nla_data(nlattrs[NL80211_ATTR_VENDOR_DATA]);
				len = nla_len(nlattrs[NL80211_ATTR_VENDOR_DATA]);

				if (resp->len > len + 8) {
					memcpy(vr->data, data, len);
					resp->len = len;
				} else
					resp->len = 0;
			}
		}
		break;
	default:
		break;
	}

	if (resp->type) {
		if (wevt->req.override_cb)
			wevt->req.override_cb(&wevt->req);
		else if (wevt->req.cb)
			wevt->req.cb(&wevt->req);
	}

	return 0;
}

static struct nlwifi_event_struct {
	const char *family;
	const char *grp;
	int (*handler)(struct nl_msg *msg, void *arg);
} nlwifi_events[] = {
	{
		.family = "nl80211",
		.grp = "scan",
		.handler = nlwifi_event_handler,
	},
	{
		.family = "nl80211",
		.grp = "config",
		.handler = nlwifi_event_handler,
	},
	{
		.family = "nl80211",
		.grp = "mlme",
		.handler = nlwifi_event_handler,
	},
	{
		.family = "nl80211",
		.grp = "vendor",
		.handler = nlwifi_event_handler,
	},
};

#define nlwifi_events_num	(sizeof(nlwifi_events)/sizeof(nlwifi_events[0]))

int nlwifi_register_event(const char *ifname, struct event_struct *req, void **handle)
{
	struct nlwifi_event *wev;
	struct nl_sock *sock;
	struct nlwifi_event_struct *ev;
	int grp;
	int err;
	int i;

	libwifi_dbg("%s %s called family %s group %s\n", ifname, __func__, req->family, req->group);

	for (i = 0; i < nlwifi_events_num; i++) {
		ev = &nlwifi_events[i];
		if (!strncmp(ev->family, req->family, strlen(ev->family)))
			break;
	}

	if (i > nlwifi_events_num) {
		libwifi_err("Error! Unknown event family '%s'\n", req->family);
		return -1;
	}

	wev = calloc(1, sizeof(struct nlwifi_event));
	if (!wev) {
		libwifi_err("%s: malloc failed!\n", __func__);
		return -1;
	}

	memcpy(&wev->req, req, sizeof(struct event_struct));
	sock = nl_socket_alloc();
	if (!sock) {
		libwifi_err("Error: nl_socket_alloc\n");
		goto free_handle;
	}
	wev->sock = sock;

	nl_socket_disable_seq_check(sock);
	err = nl_socket_modify_cb(sock, NL_CB_VALID, NL_CB_CUSTOM,
				ev->handler, wev);

	if ((err = genl_connect(sock)) < 0) {
		libwifi_err("Error: %s\n", nl_geterror(err));
		goto free_sock;
	}

	if ((grp = genl_ctrl_resolve_grp(sock, req->family, req->group)) < 0) {
		libwifi_err("Error: %s (family '%s', group '%s')\n",
					nl_geterror(grp),
					req->family,
					req->group);
		goto free_sock;
	}

	nl_socket_add_membership(sock, grp);
	req->fd_monitor = nl_socket_get_fd(sock);
	*handle = wev;

	return 0;

free_sock:
	nl_socket_free(sock);
free_handle:
	free(wev);
	*handle = NULL;
	return -1;
}

int nlwifi_unregister_event(const char *ifname, void *handle)
{
	struct nlwifi_event *wev = (struct nlwifi_event *)handle;
	struct event_struct *req;
	int ret = -1;
	int grp;

	if (!wev)
		return ret;

	req = &wev->req;
	grp = genl_ctrl_resolve_grp(wev->sock, req->family, req->group);
	if (grp < 0) {
		libwifi_err("Error: %s ('%s', '%s')\n",
					nl_geterror(grp),
					req->family,
					req->group);
		goto free_sock;
	}

	ret = 0;
	nl_socket_drop_membership(wev->sock, grp);

free_sock:
	nl_socket_free(wev->sock);
	free(wev);
	return ret;
}

int nlwifi_recv_event(const char *ifname, void *handle)
{
	struct nlwifi_event *wev = (struct nlwifi_event *)handle;
	int err;

	UNUSED(ifname);

	err = nl_recvmsgs_default(wev->sock);
	if (err < 0) {
		libwifi_err("Error: %s\n", nl_geterror(err));
		return err;
	}

	return 0;
}

const struct wifi_driver nlwifi_driver = {
	.name = "wlan",
	.scan = nlwifi_scan,
	.get_scan_results = nlwifi_get_scan_results,
	.get_bssid = nlwifi_get_bssid,
	.get_ssid = nlwifi_get_ssid,
	.get_channel = nlwifi_get_channel,
	.get_supp_channels = nlwifi_get_supp_channels,
	.get_bandwidth = nlwifi_get_bandwidth,
	.vendor_cmd = nlwifi_vendor_cmd,
	.iface.get_caps = nlwifi_ap_get_caps,
	.get_assoclist = nlwifi_get_assoclist,
	.get_sta_info = nlwifi_get_sta_info,
	.radio.get_caps = nlwifi_radio_get_caps,
	.radio.info = nlwifi_radio_info,
	.get_country = nlwifi_get_country,
	.register_event = nlwifi_register_event,
	.unregister_event = nlwifi_unregister_event,
	.recv_event = nlwifi_recv_event,
	.get_noise = nlwifi_get_noise,
	.radio.get_supp_stds = nlwifi_get_supp_stds
};
