/*
 * wpactrl_util.c - CLI to hostap and wpa_supplicant
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <net/if.h>
#include <time.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <endian.h>

#include "util.h"
#include "wifi.h"
#include "debug.h"
#include "wpactrl_util.h"

#define WPACTRL_TIMEOUT 5

/* input:
*  buf = "key1=val1\nkey2=val2\nkey3=val3"
*  param = "key2"
*  char value[x]
*  value_max = x
*  output:
*  value = "val2\0"
*/
static int get_param(const char *buf, const char *param, char *value, size_t value_max)
{
	const char *line, *loc;
	char *p, *origin_p;
	char search_str[256] = { 0 };

	if (WARN_ON(!buf) || WARN_ON(!param) || WARN_ON(!value) || WARN_ON(value_max == 0))
		return -EINVAL;

	p = strdup(buf);
	origin_p = p;

	while ((line = strsep(&p, "\n"))) {
		loc = strstr(line, param);
		if (!loc || loc != line)
			continue;
		snprintf(search_str, sizeof(search_str), "%s=%%%d[^\n]s", param, value_max);

		if (WARN_ON(sscanf(line, search_str, value) != 1)) {
			free(origin_p);
			return -1;
		}
		break;
	}
	free(origin_p);
	return 0;
}

static int hostapd_cli_get_sta_list(char *buf, uint8_t *stas, int *num_stas)
{
	const char *line;
	char *p;
	uint8_t tmp_sta[6];
	int i = 0;
	int ret;

	p = buf;
	while ((line = strsep(&p, "\n"))) {
		ret = sscanf(line, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx",
				&tmp_sta[0],
				&tmp_sta[1],
				&tmp_sta[2],
				&tmp_sta[3],
				&tmp_sta[4],
				&tmp_sta[5]);

		if (WARN_ON(ret != 6))
			continue;

		if(i >= *num_stas) {
			libwifi_err("%s Number of connected stations %d exceeds expected %d!\n",
				    __func__, (i + 1), *num_stas);
			return -EINVAL;
		}

		memcpy(&stas[i*6], tmp_sta, 6);
		i++;
	}
	*num_stas = i;
	return 0;
}

int hostapd_cli_set(const char *ifname, const char *cmd, bool check_ok)
{
	char buf[128];

	libwifi_dbg("%s %s set:\n%s\n", ifname, __func__, cmd);

	if (!ifname)
		return -1;
	if (!strlen(ifname))
		return -1;

	chrCmd(buf, sizeof(buf), "timeout %d hostapd_cli -i %s %s", WPACTRL_TIMEOUT, ifname, cmd);

	if(!check_ok)
		return 0;

	if (!strcmp(buf, "OK")) {
		libwifi_dbg("OK\n");
		return 0;
	}

	return -1;
}

int hostapd_cli_get(const char *ifname, const char *cmd, char *out, size_t out_size)
{
	if (!ifname)
		return -1;
	if (!strlen(ifname))
		return -1;

	chrCmd(out, out_size, "timeout %d hostapd_cli -i %s %s", WPACTRL_TIMEOUT, ifname, cmd);

	if (out[0] == '\0') {
		libwifi_err("%s %s failed\n", ifname, __func__);
		return -1;
	}

	libwifi_dbg("%s %s get:\n%s\n", ifname, __func__, out);
	return 0;
}

static int hostapd_cli_get_sta_num(const char *ifname, uint32_t *sta_num)
{
	const char *line;
	char buf[2048] = { 0 };
	char search_str[64] = { 0 };
	char *p;
	uint32_t num = 0;
	uint32_t tmp_num = 0;
	int i = 0;

	/* get num_sta from status */
	if (WARN_ON(hostapd_cli_get(ifname, "status", buf, sizeof(buf))))
		return -1;
	p = buf;
	while ((line = strsep(&p, "\n"))) {
		if (!strstr(line, "num_sta["))
			continue;
		snprintf(search_str, sizeof(search_str), "num_sta[%d]=%%u", i);
		i++;
		if (WARN_ON(sscanf(line, search_str, &tmp_num) != 1))
			continue;
		num += tmp_num;
	}
	*sta_num = num;
	return 0;
}

static int wpa_cli_set(const char *ifname, const char *cmd, bool check_ok)
{
	char buf[128];

	libwifi_dbg("%s %s set:\n%s\n", ifname, __func__, cmd);

	if (!ifname)
		return -1;
	if (!strlen(ifname))
		return -1;

	chrCmd(buf, sizeof(buf), "timeout %d wpa_cli -i %s %s", WPACTRL_TIMEOUT, ifname, cmd);

	if(!check_ok)
		return 0;

	if (!strcmp(buf, "OK")) {
		libwifi_dbg("OK\n");
		return 0;
	}

	return -1;
}

UNUSED_FUNC static int wpa_cli_get(const char *ifname, const char *cmd, char *out, size_t out_size)
{
	if (!ifname)
		return -1;
	if (!strlen(ifname))
		return -1;

	chrCmd(out, out_size, "timeout %d wpa_cli -i %s %s", WPACTRL_TIMEOUT, ifname, cmd);

	if (out[0] == '\0') {
		libwifi_err("%s %s failed\n", ifname, __func__);
		return -1;
	}

	libwifi_dbg("%s %s get:\n%s\n", ifname, __func__, out);
	return 0;
}

static int hostapd_cli_iface_get_sta_flags(const char *buf, char *flags, size_t flags_max)
{
	char val[256] = { 0 };
	size_t flags_len;

	if (!flags || flags_max == 0)
		return -EINVAL;

	if (get_param(buf, "flags", val, sizeof(val))) {
		libwifi_warn("%s flags parameter not found in buffer [%s]!\n",
			     __func__, buf);
		return -1;
	}

	flags_len = strlen(val);
	if (WARN_ON(flags_len >= flags_max))
		return -EINVAL;

	snprintf(flags, flags_max, "%s", val);

	return 0;
}

static int hostapd_cli_bg_rates(char *buf, uint8_t* brate, uint8_t* grate)
{
	int i;
	char val[256] = { 0 };
	char b_rates[4][3] = {
		"02",
		"04",
		"0b",
		"16"
	};
	char g_rates[8][3] = {
		"0c",
		"12",
		"18",
		"24",
		"30",
		"48",
		"60",
		"6c"
	};

	*brate = 0;
	*grate = 0;

	if (get_param(buf, "supported_rates", val, sizeof(val))) {
		libwifi_warn("%s supported_rates parameter not found in buffer [%s]!\n",
		__func__, buf);
		return -1;
	}
	if (strlen(val) < 1) {
		return -1;
	}
	for (i = 0; i < 4; i++) {
		if (!strstr(val, b_rates[i]))
			continue;
		*brate = 1;
	}
	for (i = 0; i < 8; i++) {
		if (!strstr(val, g_rates[i]))
			continue;
		*grate = 1;
	}
	return 0;
}

int hostapd_cli_iface_ap_info(const char *ifname, struct wifi_ap *ap)
{
	const char *key_mgmt = NULL;
	const char *group_cipher = NULL;
	const char *pairwise_ciphers = NULL;
	const char *bssid = NULL;
	const char *ssid = NULL;
	const char *line;
	char buf[1024];
	char *p;
	int wpa = -1;
	int ret = 0;

	UNUSED(bssid);
	UNUSED(ssid);

	ret = hostapd_cli_get(ifname, "get_config", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	libwifi_dbg("%s %s get:\n%s\n", ifname, __func__, buf);

	p = buf;
	while ((line = strsep(&p, "\n"))) {
		if (strstr(line, "wpa="))
			sscanf(line, "wpa=%d", &wpa);
		else if (strstr(line, "key_mgmt="))
			key_mgmt = line;
		else if (strstr(line, "group_cipher="))
			group_cipher = line;
		else if (strstr(line, "pairwise_cipher="))
			pairwise_ciphers = line;
		else if (strstr(line, "bssid="))
			bssid = line;
		else if (strstr(line, "ssid="))
			ssid = line;
	}

	if (strstr(key_mgmt, "WPA-PSK")) {
		switch (wpa) {
		case 1:
			ap->bss.security = BIT(WIFI_SECURITY_WPAPSK);
			break;
		case 2:
			ap->bss.security = BIT(WIFI_SECURITY_WPA2PSK);
			break;
		case 3:
			ap->bss.security = BIT(WIFI_SECURITY_WPAPSK) | BIT(WIFI_SECURITY_WPA2PSK);
			break;
		default:
			WARN_ON(1);
			break;
		}
	}


	if (pairwise_ciphers) {
		if (strstr(pairwise_ciphers, "TKIP"))
			ap->bss.rsn.pair_ciphers |= BIT(WIFI_CIPHER_TKIP);
		if (strstr(pairwise_ciphers, "CCMP"))
			ap->bss.rsn.pair_ciphers |= BIT(WIFI_CIPHER_CCMP);
	}

	if (group_cipher) {
		if (strstr(group_cipher, "TKIP"))
			ap->bss.rsn.group_cipher = WIFI_CIPHER_TKIP;
		if (strstr(group_cipher, "CCMP"))
			ap->bss.rsn.group_cipher = WIFI_CIPHER_CCMP;
	}

	libwifi_dbg("%s %s security 0x%x pair_cipers 0x%x group_cipher 0x%x\n", ifname, __func__,
		    ap->bss.security, ap->bss.rsn.pair_ciphers, ap->bss.rsn.group_cipher);

	/* get ap info from STATUS-DRIVER */
	ret = hostapd_cli_get(ifname, "raw STATUS-DRIVER", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	libwifi_dbg("%s %s STATUS-DRIVER:\n%s\n", ifname, __func__, buf);

	p = buf;
	while ((line = strsep(&p, "\n"))) {
		if (strstr(line, "capa.max_stations="))
			sscanf(line, "capa.max_stations=%d", &ap->assoclist_max);
	}

	/* get ap info from status */
	ret = hostapd_cli_get(ifname, "status", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	libwifi_dbg("%s %s hostapd status:\n%s\n", ifname, __func__, buf);

	p = buf;
	while ((line = strsep(&p, "\n"))) {
		if (strstr(line, "chan_util_avg="))
			sscanf(line, "chan_util_avg=%hhu", &ap->bss.load.utilization);
	}

	return ret;
}

int hostapd_cli_iface_get_assoclist(const char *ifname, uint8_t *stas, int *num_stas)
{
	char buf[1024];
	uint32_t num = 0;
	int tmp_num = *num_stas;

	if (!stas || *num_stas <= 0) {
		libwifi_err("[%s] %s Invalid args!\n", ifname, __func__);
		return -EINVAL;
	}

	if (hostapd_cli_get_sta_num(ifname, &num))
		return -1;

	if (num == 0) {
		*num_stas = num;
		return 0;
	}

	if (WARN_ON(num >= *num_stas)) {
		libwifi_err("[%s] %s Number of connected stations %d exceeds expected %d!\n",
			ifname, __func__, num, *num_stas);
		return -EINVAL;
	}

	if (hostapd_cli_get(ifname, "list_sta", buf, sizeof(buf)))
		return -1;

	if (hostapd_cli_get_sta_list(buf, stas, &tmp_num))
		return -1;

	if (tmp_num != num) {
		libwifi_warn("[%s] %s Number of connected stations %d differs from list %d!\n",
			ifname, __func__, num, tmp_num);
		num = tmp_num;
	}

	*num_stas = num;
	return 0;
}

static int hostapd_cli_get_sta_capability_information(const char *sta_buf, uint8_t *ie, size_t *ie_len) {
	char val[256] = { 0 };
	uint16_t capa = 0;
	uint8_t *ptr;

	if (WARN_ON(!ie) || WARN_ON(*ie_len < 2))
		return -EINVAL;

	if (get_param(sta_buf, "capability", val, sizeof(val))) {
		libwifi_warn("%s capability parameter not found in buffer [%s]!\n",
		__func__, sta_buf);
		return -1;
	}

	if (sscanf(val, "%hx", &capa) != 1)
		return -1;

	ptr = (uint8_t *)&capa;
	ie[0] = ptr[0];
	ie[1] = ptr[1];
	*ie_len = 2;

	return 0;
}

static int hostapd_cli_get_sta_ht_capabilities_info(const char *sta_buf, uint8_t *ie, size_t *ie_len) {
	char val[256] = { 0 };
	uint16_t capa = 0;
	uint8_t *ptr;

	if (WARN_ON(!ie) || WARN_ON(*ie_len < 2))
		return -EINVAL;

	if (get_param(sta_buf, "ht_caps_info", val, sizeof(val))) {
		libwifi_warn("%s ht_caps_info parameter not found in buffer [%s]!\n",
		__func__, sta_buf);
		return -1;
	}

	if (sscanf(val, "%hx", &capa) != 1)
		return -1;

	ptr = (uint8_t *)&capa;
	ie[0] = ptr[0];
	ie[1] = ptr[1];
	*ie_len = 2;

	return 0;
}

static int hostapd_cli_get_sta_vht_capabilities_info(const char *sta_buf, uint8_t *ie, size_t *ie_len) {
	char val[256] = { 0 };
	uint32_t capa = 0;
	uint8_t *ptr;

	if (WARN_ON(!ie) || WARN_ON(*ie_len < 4))
		return -EINVAL;

	if (get_param(sta_buf, "vht_caps_info", val, sizeof(val))) {
		libwifi_warn("%s vht_caps_info parameter not found in buffer [%s]!\n",
		__func__, sta_buf);
		return -1;
	}

	if (sscanf(val, "%x", &capa) != 1)
		return -1;
	ptr = (uint8_t *)&capa;
	ie[0] = ptr[0];
	ie[1] = ptr[1];
	ie[2] = ptr[2];
	ie[3] = ptr[3];
	*ie_len = 4;

	return 0;
}

static int hostapd_cli_get_extended_capabilities(const char *sta_buf, uint8_t *ie, size_t *ie_len)
{
	int i = 0;
	size_t ext_len;
	char val[256] = { 0 };
	uint8_t octet = 0;
	char *ptr;

	if (WARN_ON(!ie) || WARN_ON(*ie_len < 8))
		return -EINVAL;

	if (get_param(sta_buf, "ext_capab", val, sizeof(val))) {
		libwifi_warn("%s ext_capab parameter not found in buffer [%s]!\n",
		__func__, sta_buf);
		return -1;
	}
	ext_len = strlen(val);

	if (ext_len == 0 || ext_len % 2 != 0) {
                libwifi_err("Invalid format of ext_capab[%s] len[%ld]\n", val, ext_len);
                return -1;
        }

	ptr = val;
	while (*ptr != '\0' && i < *ie_len) {
		sscanf(ptr, "%2hhx", &octet);
		ie[i] = octet;
		ptr += 2;
		i++;
        }
	*ie_len = i;

	return 0;
}

int hostapd_cli_iface_get_sta_info(const char *ifname, uint8_t *addr, struct wifi_sta *info)
{
	char sta_flags[128] = { 0 };
	char cmd[64] = { 0 };
	char buf[2048] = { 0 };
	uint8_t ie[128] = { 0 };
	int ret = 0;
	size_t ie_max = sizeof(ie);
	size_t ie_len;

	libwifi_dbg("%s %s \n", ifname, __func__);

	snprintf(cmd, sizeof(cmd), "sta " MACSTR, MAC2STR(addr));

	if (WARN_ON(hostapd_cli_get(ifname, cmd, buf, sizeof(buf))))
		return -1;

	ie_len = ie_max;
	ret = hostapd_cli_get_sta_capability_information(buf, ie, &ie_len);

	if (WARN_ON(ret))
		wifi_cap_set_from_capability_information(info->cbitmap, ie, ie_len);

	hostapd_cli_iface_get_sta_flags(buf, sta_flags, sizeof(sta_flags));

	if (strstr(sta_flags, "[HT]")) {
		info->caps.valid |= WIFI_CAP_HT_VALID;
		ie_len = ie_max;
		ret = hostapd_cli_get_sta_ht_capabilities_info(buf, ie, &ie_len);
		if (!ret)
			wifi_cap_set_from_ht_capabilities_info(info->cbitmap, ie, ie_len);
	}

	if (strstr(sta_flags, "[VHT]")) {
		info->caps.valid |= WIFI_CAP_VHT_VALID;
		ie_len = ie_max;
		ret = hostapd_cli_get_sta_vht_capabilities_info(buf, ie, &ie_len);
		if (!ret)
			wifi_cap_set_from_vht_capabilities_info(info->cbitmap, ie, ie_len);
	}

	ie_len = ie_max;
	ret = hostapd_cli_get_extended_capabilities(buf, ie, &ie_len);
	if (!ret) {
		info->caps.valid |= WIFI_CAP_EXT_VALID;
		wifi_cap_set_from_extended_capabilities(info->cbitmap, ie, ie_len);
	}

	return ret;
}

int hostapd_cli_get_ssid(const char *ifname, char *ssid, size_t ssid_size)
{
	char buf[1024];
	int ret;

	memset(ssid, 0, ssid_size);

	ret = hostapd_cli_get(ifname, "get_config | grep ^ssid", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	snprintf(ssid, ssid_size, "%s", buf + 5);

	return ret;
}

int hostapd_cli_iface_add_neighbor(const char *ifname, struct nbr *nbr, size_t nbr_size)
{
	char bssid[18] = { 0 };
	char nr[256] = { 0 };
	char cmd[1024] = { 0 };
	char ssid[128];

	if (hostapd_cli_get_ssid(ifname, ssid, sizeof(ssid)))
		return -1;

	snprintf(bssid, sizeof(bssid), MACSTR, MAC2STR(nbr->bssid));
	btostr((uint8_t *) nbr, nbr_size, nr);

	snprintf(cmd, sizeof(cmd), "set_neighbor " MACSTR " ssid=\\\"%s\\\" nr=%s",
		 MAC2STR(nbr->bssid), ssid, nr);

	return hostapd_cli_set(ifname, cmd, true);
}

int hostapd_cli_iface_del_neighbor(const char *ifname, unsigned char *bssid)
{
	char cmd[1024] = { 0 };
	char ssid[128];

	if (hostapd_cli_get_ssid(ifname, ssid, sizeof(ssid)))
		return -1;

	snprintf(cmd, sizeof(cmd), "remove_neighbor " MACSTR " ssid=\\\"%s\\\"",
		 MAC2STR(bssid), ssid);

	return hostapd_cli_set(ifname, cmd, true);
}

int hostapd_cli_iface_start_wps(const char *ifname, struct wps_param wps)
{
	char cmd[1024] = { 0 };
	enum wifi_mode mode;
	int ret = 0;

	if (wps.method == WPS_METHOD_PBC){
		snprintf(cmd, sizeof(cmd),"wps_pbc");
	} else if (wps.method == WPS_METHOD_PIN) {
		snprintf(cmd, sizeof(cmd),"wps_pin %s %08lu", "any", wps.pin);
	}

	ret = wifi_get_mode(ifname, &mode);

	if(ret != 0)
		return ret;

	switch (mode) {
	case WIFI_MODE_AP:
	case WIFI_MODE_AP_VLAN:
		ret = hostapd_cli_set(ifname, cmd, true);
		break;
	case WIFI_MODE_STA:
		ret = wpa_cli_set(ifname, cmd, false); //Do not check OK answer. wpa_cli wps_pin returns pin.
		break;
	default:
		WARN_ON(1);
		ret = -1;
		break;
	}

	return ret;
}

int hostapd_cli_iface_stop_wps(const char *ifname)
{
	enum wifi_mode mode;
	int ret = 0;

	ret = wifi_get_mode(ifname, &mode);

	if(ret != 0)
		return ret;

	switch (mode) {
	case WIFI_MODE_AP:
	case WIFI_MODE_AP_VLAN:
		ret = hostapd_cli_set(ifname, "wps_cancel", true);
		break;
	case WIFI_MODE_STA:
		ret = wpa_cli_set(ifname, "wps_cancel", true);
		break;
	default:
		WARN_ON(1);
		ret = -1;
		break;
	}

	return ret;
}

int hostapd_cli_iface_get_wps_status(const char *ifname, enum wps_status *s)
{
	char buf[256] = {0};
	int ret;

	ret = hostapd_cli_get(ifname, "wps_get_status", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	const char *f1 = "PBC Status: ";
	const char *f2 = "Last WPS result: ";
	char *pos1, *pos2;

	pos1 = strstr(buf, f1);
	pos2 = strstr(buf, f2);

	if (!pos1 || !pos2)
		return -1;

	pos1 += strlen(f1);
	if (strstr(pos1, "Disabled"))
		*s = WPS_STATUS_INIT;
	else if (strstr(pos1, "Active"))
		*s = WPS_STATUS_PROCESSING;
	else if (strstr(pos1, "Timed-out"))
		*s = WPS_STATUS_TIMEOUT;
	else if (strstr(pos1, "Overlap"))
		*s = WPS_STATUS_FAIL;
	else
		*s = WPS_STATUS_UNKNOWN;

	if (strstr(pos2, "Success"))
		*s = WPS_STATUS_SUCCESS;

	libwifi_dbg("%s %s status %d \n", ifname, __func__,*s);
	return 0;
}

int hostapd_cli_iface_get_wps_ap_pin(const char *ifname, unsigned long *pin)
{
	char buf[256] = {0};
	int ret = 0;

	ret = hostapd_cli_get(ifname, "wps_ap_pin get", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	if (WARN_ON(1 != sscanf(buf,"%lu",pin)))
		ret = -1;

	return ret;
}

int hostapd_cli_iface_set_wps_ap_pin(const char *ifname, unsigned long pin)
{
	char cmd[1024] = {0};
	char buf[256] = {0};
	char pinbuf[9] = {0};
	int ret = 0;

	snprintf(pinbuf, 9, "%08lu", pin);
	snprintf(cmd, sizeof(cmd),"wps_ap_pin set %08lu", pin);

	ret = hostapd_cli_set(ifname, cmd, false);
	if (WARN_ON(ret))
		return ret;

	ret = hostapd_cli_get(ifname, "wps_ap_pin get", buf, sizeof(buf));
	if (WARN_ON(ret))
		return ret;

	if (WARN_ON(!strstr(buf, pinbuf)))
		ret = -1;

	return ret;
}

static int hostapd_cli_iface_get_neighbor_num(const char *ifname)
{
	char buf[512];

	if (WARN_ON(hostapd_cli_get(ifname, "show_neighbor | wc -l", buf, sizeof(buf))))
		return 0;

	return atoi(buf);
}

int hostapd_cli_iface_get_neighbor_list(const char *ifname, struct nbr *nbrs, int *nr)
{
	struct nbr *nbr;
	char *p, *k, *l;
	char *line;
	char *word;
	char *buf;
	int count;
	int len;
	int ret;

	/* make sure we have place for show_neighbor output */
	buf = calloc(*nr, 256);
	if (WARN_ON(!buf))
		return -1;

	ret = hostapd_cli_get(ifname, "show_neighbor", buf, *nr * 256);
	if (WARN_ON(ret)) {
		free(buf);
		return ret;
	}

	nbr = nbrs;
	count = 0;

	p = buf;
	while ((line = strsep(&p, "\n"))) {
		k = line;
		l = NULL;

		if (WARN_ON(count >= *nr))
			break;

		while ((word = strsep(&k, " "))) {
			if (!(l = strstr(word, "nr=")))
				continue;

			len = strlen(l) - 3;
			if (len > 2 * sizeof(*nbr))
				len = 2 * sizeof(*nbr);

			strtob(l + 3, len, (uint8_t *) nbr);
			break;
		}

		if (!l)
			continue;

		libwifi_dbg("nbr[%d] " MACSTR " bssid_info 0x%x reg %u chan %u phy %u\n",
			    count, MAC2STR(nbr->bssid), nbr->bssid_info, nbr->reg,
			    nbr->channel, nbr->phy);

		count++;
		nbr = &nbrs[count];
	}

	free(buf);
	*nr = count;
	return ret;
}

int hostapd_cli_iface_req_beacon(const char *ifname, unsigned char *sta,
				 struct rrm_measurement_beacon_request *req,
				 size_t req_size)
{
	char cmd[1024] = { 0 };
	char req_hex[256] = { 0 };
	size_t size;

	size = req_size;
	if (WARN_ON(req_size * 2 >= sizeof(req_hex)))
		size = sizeof(req_hex) / 2;

	/* convert to LE */
	req->rand_interval = htole16(req->rand_interval);
	req->duration = htole16(req->duration);

	btostr((uint8_t *) req, size, req_hex);
	snprintf(cmd, sizeof(cmd), "req_beacon " MACSTR " %s",
		 MAC2STR(sta), req_hex);

	/* For some reason req_beacon don't reply with OK */
	hostapd_cli_set(ifname, cmd, false);
	return 0;
}

int hostapd_cli_disconnect_sta(const char *ifname, uint8_t *sta, uint16_t reason)
{
	char cmd[1024] = { 0 };

	snprintf(cmd, sizeof(cmd), "deauthenticate " MACSTR, MAC2STR(sta));

	if (reason > 0)
		snprintf(cmd + strlen(cmd), sizeof(cmd), " reason=%d", reason);

	return hostapd_cli_set(ifname, cmd, true);
}

int hostapd_cli_sta_disconnect_ap(const char *ifname, uint32_t reason)
{
	/* reason parameter is skiped because wpa_supplicant command
	 * disconnect does not allow setup this parameter,
	 * it is hardcoded to 3.
	 */
	return wpa_cli_set(ifname, "disconnect", true);
}

int hostapd_cli_iface_req_bss_transition(const char *ifname, unsigned char *sta,
					 int bsss_nr, unsigned char *bsss,
					 struct bss_transition_params *params)
{
	unsigned char *bss;
	struct nbr *nbrs;
	struct nbr *nbr;
	size_t buflen;
	size_t len;
	int i, j, nr;
	char *buf;
	int ret;

	nr = hostapd_cli_iface_get_neighbor_num(ifname);
	if (WARN_ON(bsss_nr && !nr))
		return -1;

	nbrs = calloc(nr, sizeof(*nbr));
	if (WARN_ON(!nbrs))
		return -1;

	if (hostapd_cli_iface_get_neighbor_list(ifname, nbrs, &nr)) {
		free(nbrs);
		return -1;
	}

	buflen = 256 + bsss_nr * 256;

	buf = malloc(buflen);
	if (WARN_ON(!buf)) {
		free(nbrs);
		return -1;
	}

	len = 0;
	len += snprintf(buf + len, buflen - len, "bss_tm_req " MACSTR " ", MAC2STR(sta));

	for (i = 0 ; i < bsss_nr; i++) {
		bss = bsss + i * 6;

		/* Check neighbor list */
		for (j = 0; j < nr; j++) {
			if (!memcmp(bss, nbrs[j].bssid, 6))
				break;
		}

		if (j == nr) {
			libwifi_dbg("skip " MACSTR " - not on neighbor list\n", MAC2STR(bss));
			continue;
		}

		nbr = &nbrs[j];
		len += snprintf(buf + len, buflen - len,
				"neighbor=" MACSTR ",%u,%hhu,%hhu,%hhu ",
				MAC2STR(nbr->bssid), nbr->bssid_info, nbr->reg,
				nbr->channel, nbr->phy);
	}

	len += snprintf(buf + len, buflen - len,
			"pref=%hhu disassoc_imminent=%hhu ",
			params->pref, params->disassoc_imminent);

	if (params->valid_int)
		len += snprintf(buf + len, buflen - len,
				"valid_int=%hu", params->valid_int);

	if (params->bss_term)
		len += snprintf(buf + len, buflen - len,
				"bss_term=%hu ", params->bss_term);
	if (params->disassoc_timer)
		len += snprintf(buf + len, buflen - len,
				"disassoc_time=%hu ", params->disassoc_timer);
	if (params->abridged)
		len += snprintf(buf + len, buflen - len,
				"abridged=%hhu ", params->abridged);

	ret = hostapd_cli_set(ifname, buf, false);

	free(nbrs);
	free(buf);

	return ret;
}

int hostapd_cli_iface_add_vendor_ie(const char *ifname, struct vendor_iereq *req)
{
	char cmd[1024] = { 0 };
	char ie[512] = { 0 };
	size_t size;

	if (WARN_ON(!req->mgmt_subtype))
		return 0;

	size = req->ie.ie_hdr.len + sizeof(req->ie.ie_hdr);
	if (WARN_ON(size * 2 >= sizeof(ie)))
		return -1;

	btostr((uint8_t *) &req->ie, size, ie);
	snprintf(cmd, sizeof(cmd), "set vendor_elements %s", ie);

	return hostapd_cli_set(ifname, cmd, true);
}

int hostapd_cli_iface_del_vendor_ie(const char *ifname, struct vendor_iereq *req)
{
	char cmd[1024] = { 0 };

	snprintf(cmd, sizeof(cmd), "set vendor_elements \"\"");
	return hostapd_cli_set(ifname, cmd, true);
}

int hostapd_cli_is_wds_sta(const char *ifname, uint8_t *sta_mac, char *ifname_wds, size_t ifname_max)
{
	char cmd[64] = { 0 };
	char val[256] = { 0 };
	char buf[2048] = { 0 };
	int is_wds = 0;

	snprintf(cmd, sizeof(cmd), "sta " MACSTR, MAC2STR(sta_mac));
	if (hostapd_cli_get(ifname, cmd, buf, sizeof(buf)))
		return 0;

	if (hostapd_cli_iface_get_sta_flags(buf, val, sizeof(val)))
		return 0;

	if (strstr(val, "[WDS]") || strstr(val, "[MULTI_AP]"))
		is_wds = 1;

	if (!ifname_wds)
		return is_wds;

	if (get_param(buf, "ifname_wds", val, sizeof(val))) {
		libwifi_warn("[%s] %s ifname_wds parameter not found in buffer [%s]!\n",
			    ifname, __func__, buf);
		return is_wds;
	}

	if (WARN_ON((strlen(val) + 1) > ifname_max))
		return is_wds;

	snprintf(ifname_wds, ifname_max, "%s", val);

	return is_wds;
}

int hostapd_cli_get_oper_stds(const char *ifname, uint8_t *std)
{
	char buf[2048] = { 0 };
	char val[256] = { 0 };
	uint8_t n_state;
	uint8_t ac_state;
	uint8_t channel;
	uint8_t brate, grate;

	if (WARN_ON(hostapd_cli_get(ifname, "status", buf, sizeof(buf))))
		return -1;

	if (get_param(buf, "channel", val, sizeof(val))) {
		libwifi_warn("%s channel parameter not found in buffer [%s]!\n",
		__func__, buf);
	}
	if (WARN_ON(sscanf(val, "%hhd", &channel) != 1)) {
		libwifi_warn("Failed to parse channel parameter\n");
	}
	if (get_param(buf, "ieee80211n", val, sizeof(val))) {
		libwifi_warn("%s ieee80211n parameter not found in buffer [%s]!\n",
		__func__, buf);
	}
	if (WARN_ON(sscanf(val, "%1hhd", &n_state) != 1)) {
		libwifi_warn("Failed to parse ieee80211n parameter\n");
	}
	if (get_param(buf, "ieee80211ac", val, sizeof(val))) {
		libwifi_warn("%s ieee80211ac parameter not found in buffer [%s]!\n",
		__func__, buf);
	}
	if (WARN_ON(sscanf(val, "%1hhd", &ac_state) != 1)) {
		libwifi_warn("Failed to parse ieee80211ac parameter\n");
	}

	*std = 0;

	if (ac_state) {
		*std |= WIFI_AC;
	}

	if (n_state) {
		*std |= WIFI_N;
		if (channel >= 36)
			*std |= WIFI_A;
		return 0;
	}

	if (channel >= 36) {
		*std |= WIFI_A;
		return 0;
	}

	hostapd_cli_bg_rates(buf, &brate, &grate);

	if (brate)
		*std |= WIFI_B;
	if (grate)
		*std |= WIFI_G;

	return 0;
}

int hostapd_ubus_iface_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta_config *cfg)
{
	char buf[256];

	libwifi_dbg("[%s] hostapd monitor sta " MACSTR " %s\n", ifname, MAC2STR(sta),
		    cfg->enable ? "enable" : "disable");

	/* Today we enable probe-req monitor in hostapd - even STA isn't connected */
	chrCmd(buf, sizeof(buf), "ubus call hostapd.%s %s '{\"addr\":\"" MACSTR "\"}'",
	       ifname, cfg->enable ? "sta_monitor_add" : "sta_monitor_del", MAC2STR(sta));

	if (cfg->enable)
		chrCmd(buf, sizeof(buf),
		       "ubus call hostapd.%s event_mask_add '{\"event\":\"probe-req-filtered\"}'",
		       ifname);

        return 0;
}

int hostapd_ubus_iface_get_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta *mon)
{
	char buf[1024];
	char format[1024];
	int rssi;
	int age;

	libwifi_dbg("[%s] get monitor sta " MACSTR "\n", ifname, MAC2STR(sta));

	chrCmd(buf, sizeof(buf), "ubus -S call hostapd.%s sta_monitor_get '{\"addr\":\"" MACSTR "\"}'",
	       ifname, MAC2STR(sta));

	snprintf(format, sizeof(format), "{\"sta\":{\"macaddr\":\"" MACSTR "\",\"rssi\":%%d,\"age\":%%d}}", MAC2STR(sta));
	/* Format: {"sta":{"macaddr":"30:10:b3:6d:8d:ba","rssi":-29,"age":782}} */
	if (WARN_ON(sscanf(buf, format, &rssi, &age) != 2))
		return -1;

	memcpy(mon->macaddr, sta, 6);
	mon->rssi_avg = rssi;
	mon->last_seen = age;

	return 0;
}

int hostapd_ubus_iface_get_monitor_stas(const char *ifname, struct wifi_monsta *stas, int *num)
{
	char buf[1024];
	struct wifi_monsta *sta;
	char *line;
	char *key;
	char *p;
	int i;

	chrCmd(buf, sizeof(buf), "ubus call hostapd.%s sta_monitor_get", ifname);

	p = buf;
	i = 0;

	while ((line = strsep(&p, "\n"))) {
		if (WARN_ON(i >= *num))
			break;

		key = strstr(line, "\"macaddr\":");
		if (!key)
			continue;

		sta = &stas[i];
		if (sscanf(key, "\"macaddr\": \"%hhx:%hhx:%hhx:%hhx:%hhx:%hhx\"",
			   &sta->macaddr[0],
			   &sta->macaddr[1],
			   &sta->macaddr[2],
			   &sta->macaddr[3],
			   &sta->macaddr[4],
			   &sta->macaddr[5]) != 6)
			continue;
		i++;
	}

	*num = i;
	return 0;
}

static int
hostapd_ubus_iface_subscribe_unsubscribe_frame(const char *ifname, uint8_t type, uint8_t stype, int sub)
{
	char buf[1024];
	char evt[64];
	char *res;

	if (type != WIFI_FRAME_MGMT)
		return -ENOTSUP;

	switch (stype) {
	case WIFI_FRAME_ASSOC_REQ:
		snprintf(evt, sizeof(evt), "raw-assoc");
		break;
	case WIFI_FRAME_REASSOC_REQ:
		snprintf(evt, sizeof(evt), "raw-reassoc");
		break;
	case WIFI_FRAME_PROBE_REQ:
		snprintf(evt, sizeof(evt), "raw-probe-req");
		break;
	case WIFI_FRAME_DISASSOC:
		snprintf(evt, sizeof(evt), "raw-disassoc");
		break;
	case WIFI_FRAME_AUTH:
		snprintf(evt, sizeof(evt), "raw-auth");
		break;
	case WIFI_FRAME_DEAUTH:
		snprintf(evt, sizeof(evt), "raw-deauth");
		break;
	case WIFI_FRAME_ACTION:
		snprintf(evt, sizeof(evt), "raw-action");
		break;
	default:
		return -ENOTSUP;
	}

	chrCmd(buf, sizeof(buf), "ubus call hostapd.%s %s '{\"event\":\"%s\"}'",
	       ifname, sub ? "event_mask_add" : "event_mask_del", evt);

	chrCmd(buf, sizeof(buf), "ubus -S call hostapd.%s event_mask_get", ifname);
	res = strstr(buf, evt);

	if (sub && !res)
		return -1;

	if (!sub && res)
		return -1;

	return 0;
}

int hostapd_ubus_iface_subscribe_frame(const char *ifname, uint8_t type, uint8_t stype)
{
	return hostapd_ubus_iface_subscribe_unsubscribe_frame(ifname, type, stype, 1);
}

int hostapd_ubus_iface_unsubscribe_frame(const char *ifname, uint8_t type, uint8_t stype)
{
	return hostapd_ubus_iface_subscribe_unsubscribe_frame(ifname, type, stype, 0);
}

int hostapd_cli_iface_chan_switch(const char *ifname, struct chan_switch_param *param)
{
	char cmd[1024] = { 0 };

	/* <cs_count> <freq> [sec_channel_offset=] [center_freq1=] [center_freq2=] [bandwidth=] [blocktx] [ht|vht] */
	snprintf(cmd, sizeof(cmd), "chan_switch %d %d", param->count, param->freq);
	if (param->sec_chan_offset)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " sec_channel_offset=%d", param->sec_chan_offset);
	if (param->cf1)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " center_freq1=%d", param->cf1);
	if (param->cf2)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " center_freq2=%d", param->cf2);
	if (param->bandwidth)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " bandwidth=%d", param->bandwidth);
	if (param->blocktx)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " blocktx");
	if (param->vht)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " vht");
	else if (param->ht)
		snprintf(cmd + strlen(cmd), sizeof(cmd) - strlen(cmd), " ht");

	return hostapd_cli_set(ifname, cmd, true);
}
